@@ -10,6 +10,7 @@
 #include <errno.h>
 #include <stdbool.h>
 #include <stddef.h>
+#include <stdint.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -103,13 +104,17 @@ EXPORT_SYMBOL hxmc_t *HX_getl(hxmc_t **ptr, FILE *fp)
 	if (fgets(temp, sizeof(temp), fp) == NULL)
 		return NULL;
 
-	if (*ptr == NULL)
+	if (*ptr == NULL) {
 		*ptr = HXmc_meminit(NULL, 0);
-	else
+		if (*ptr == NULL)
+			return NULL;
+	} else {
 		HXmc_trunc(ptr, 0);
+	}
 
 	do {
-		HXmc_strcat(ptr, temp);
+		if (HXmc_strcat(ptr, temp) == NULL)
+			return *ptr;
 		if (strchr(temp, '\n') != NULL)
 			break;
 	} while (fgets(temp, sizeof(temp), fp) != NULL);
@@ -253,6 +258,39 @@ EXPORT_SYMBOL char *HX_strclone(char **pa, const char *pb)
 	return *pa;
 }
 
+EXPORT_SYMBOL char *HX_strdup(const char *src)
+{
+	if (src == NULL)
+		return NULL;
+	/* return HX_strndup(src, SIZE_MAX); */
+	return HX_memdup(src, strlen(src) + 1);
+}
+
+EXPORT_SYMBOL char *HX_strlcat(char *dest, const char *src, size_t len)
+{
+	ssize_t x = len - strlen(dest) - 1;
+	if (x <= 0)
+		return dest;
+	return strncat(dest, src, x);
+}
+
+EXPORT_SYMBOL char *HX_strlcpy(char *dest, const char *src, size_t n)
+{
+	strncpy(dest, src, n);
+	dest[n-1] = '\0';
+	return dest;
+}
+
+EXPORT_SYMBOL char *HX_strlncat(char *dest, const char *src, size_t dlen,
+    size_t slen)
+{
+	ssize_t x = dlen - strlen(dest) - 1;
+	if (x <= 0)
+		return dest;
+	x = ((ssize_t)slen < x) ? (ssize_t)slen : x;
+	return strncat(dest, src, x);
+}
+
 EXPORT_SYMBOL char *HX_strlower(char *orig)
 {
 	char *expr;
@@ -290,6 +328,31 @@ EXPORT_SYMBOL char *HX_strmid(const char *expr, long offset, long length)
 	return HX_strlcpy(buffer, expr, length + 1);
 }
 
+EXPORT_SYMBOL char *HX_strndup(const char *src, size_t size)
+{
+	char *ret;
+	size_t z;
+
+	if (src == NULL)
+		return NULL;
+	z = strlen(src);
+	if (z < size)
+		size = z;
+	if ((ret = malloc(size + 1)) == NULL)
+		return NULL;
+	memcpy(ret, src, size);
+	ret[size] = '\0';
+	return ret;
+}
+
+EXPORT_SYMBOL size_t HX_strnlen(const char *src, size_t size)
+{
+	const char *ptr = src;
+	for (; *ptr != '\0' && size > 0; --size, ++ptr)
+		;
+	return ptr - src;
+}
+
 EXPORT_SYMBOL size_t HX_strrcspn(const char *s, const char *rej)
 {
 	size_t n = strlen(s);
@@ -418,6 +481,45 @@ static char *HX_quote_backslash(char *dest, const char *src, const char *qc)
 	return ret;
 }
 
+/**
+ * Encode @src into BASE-64 according to RFC 4648 and write result to @dest,
+ * which must be of appropriate size, plus one for a trailing NUL.
+ */
+static char *HX_quote_base64(char *d, const char *s)
+{
+	static const char a[] =
+		"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
+		"abcdefghijklmnopqrstuvwxyz0123456789+/";
+	size_t len = strlen(s);
+	char *ret = d;
+
+	while (len > 0) {
+		if (len >= 3) {
+			len -= 3;
+			d[0] = a[(s[0] & 0xFC) >> 2];
+			d[1] = a[((s[0] & 0x03) << 4) | ((s[1] & 0xF0) >> 4)];
+			d[2] = a[((s[1] & 0x0F) << 2) | ((s[2] & 0xC0) >> 6)];
+			d[3] = a[s[2] & 0x3F];
+		} else if (len == 2) {
+			len = 0;
+			d[0] = a[(s[0] & 0xFC) >> 2];
+			d[1] = a[((s[0] & 0x03) << 4) | ((s[1] & 0xF0) >> 4)];
+			d[2] = a[(s[1] & 0x0F) << 2];
+			d[3] = '=';
+		} else if (len == 1) {
+			len = 0;
+			d[0] = a[(s[0] & 0xFC) >> 2];
+			d[1] = a[(s[0] & 0x03) << 4];
+			d[2] = '=';
+			d[3] = '=';
+		}
+		s += 3;
+		d += 4;
+	}
+	*d = '\0';
+	return ret;
+}
+
 static size_t HX_qsize_html(const char *s)
 {
 	const char *p = s;
@@ -515,6 +617,8 @@ static size_t HX_quoted_size(const char *s, unsigned int type)
 	case HXQUOTE_LDAPFLT:
 	case HXQUOTE_LDAPRDN:
 		return HX_qsize_backslash(s, HX_quote_chars[type], 2);
+	case HXQUOTE_BASE64:
+		return (strlen(s) + 2) / 3 * 4;
 	default:
 		return strlen(s);
 	}
@@ -526,7 +630,13 @@ EXPORT_SYMBOL char *HX_strquote(const char *src, unsigned int type,
 	bool do_quote;
 	char *tmp;
 
-	do_quote = type < _HXQUOTE_MAX &&
+	if (type >= _HXQUOTE_MAX) {
+		errno = EINVAL;
+		return NULL;
+	}
+	/* If quote_chars is NULL, it is clear all chars are to be encoded. */
+	do_quote = type >= ARRAY_SIZE(HX_quote_chars) ||
+	           HX_quote_chars[type] == NULL ||
 	           strpbrk(src, HX_quote_chars[type]) != NULL;
 	/*
 	 * free_me == NULL implies that we always allocate, even if
@@ -556,6 +666,8 @@ EXPORT_SYMBOL char *HX_strquote(const char *src, unsigned int type,
 	case HXQUOTE_LDAPFLT:
 	case HXQUOTE_LDAPRDN:
 		return HX_quote_ldap(*free_me, src, HX_quote_chars[type]);
+	case HXQUOTE_BASE64:
+		return HX_quote_base64(*free_me, src);
 	}
 	return NULL;
 }
